package solutions.bitbadger.documents.common

/**
 * A field and its comparison
 *
 * @property name The name of the field in the JSON document
 * @property comparison The comparison to apply against the field
 * @property parameterName The name of the parameter to use in the query (optional, generated if missing)
 * @property qualifier A table qualifier to use to address the `data` field (useful for multi-table queries)
 */
class Field<T>(
    val name: String,
    val comparison: Comparison<T>,
    val parameterName: String? = null,
    val qualifier: String? = null) {

    /**
     * Get the path for this field
     *
     * @param dialect The SQL dialect to use for the path to the JSON field
     * @param format Whether the value should be retrieved as JSON or SQL (optional, default SQL)
     * @return The path for the field
     */
    fun path(dialect: Dialect, format: FieldFormat = FieldFormat.SQL): String =
        (if (qualifier == null) "" else "${qualifier}.") + nameToPath(name, dialect, format)

    companion object {
        /**
         * Create a field equality comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> Equal(name: String, value: T): Field<T> =
            Field<T>(name, Comparison(Op.EQUAL, value))

        /**
         * Create a field greater-than comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> Greater(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.GREATER, value))

        /**
         * Create a field greater-than-or-equal-to comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> GreaterOrEqual(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.GREATER_OR_EQUAL, value))

        /**
         * Create a field less-than comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> Less(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.LESS, value))

        /**
         * Create a field less-than-or-equal-to comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> LessOrEqual(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.LESS_OR_EQUAL, value))

        /**
         * Create a field inequality comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> NotEqual(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.NOT_EQUAL, value))

        /**
         * Create a field range comparison
         *
         * @param name The name of the field to be compared
         * @param minValue The lower value for the comparison
         * @param maxValue The upper value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> Between(name: String, minValue: T, maxValue: T): Field<Pair<T, T>> =
            Field(name, Comparison(Op.BETWEEN, Pair(minValue, maxValue)))

        fun <T> In(name: String, values: List<T>): Field<List<T>> =
            Field(name, Comparison(Op.IN, values))

        fun <T> InArray(name: String, tableName: String, values: List<T>): Field<Pair<String, List<T>>> =
            Field(name, Comparison(Op.IN_ARRAY, Pair(tableName, values)))

        fun Exists(name: String): Field<String> =
            Field(name, Comparison(Op.EXISTS, ""))

        fun NotExists(name: String): Field<String> =
            Field(name, Comparison(Op.NOT_EXISTS, ""))

        fun Named(name: String): Field<String> =
            Field(name, Comparison(Op.EQUAL, ""))

        fun nameToPath(name: String, dialect: Dialect, format: FieldFormat): String {
            val path = StringBuilder("data")
            val extra = if (format == FieldFormat.SQL) ">" else ""
            if (name.indexOf('.') > -1) {
                if (dialect == Dialect.POSTGRESQL) {
                    path.append("#>", extra, "'{", name.replace('.', ','), "}'")
                } else {
                    val names = mutableListOf(name.split('.'))
                    val last = names.removeLast()
                    names.forEach { path.append("->'", it, "'") }
                    path.append("->", extra, "'", last, "'")
                }
            } else {
                path.append("->", extra, "'", name, "'")
            }
            return path.toString()
        }
    }
}