package solutions.bitbadger.documents.common

object Query {

    /**
     * Combine a query (`SELECT`, `UPDATE`, etc.) and a `WHERE` clause
     *
     * @param statement The first part of the statement
     * @param where The `WHERE` clause for the statement
     * @return The two parts of the query combined with `WHERE`
     */
    fun statementWhere(statement: String, where: String): String =
        "$statement WHERE $where"

    object Definition {

        /**
         * SQL statement to create a document table
         *
         * @param tableName The name of the table to create (may include schema)
         * @param dataType The type of data for the column (`JSON`, `JSONB`, etc.)
         * @return A query to create a document table
         */
        fun ensureTableFor(tableName: String, dataType: String): String =
            "CREATE TABLE IF NOT EXISTS $tableName (data $dataType NOT NULL)"

        /**
         * Split a schema and table name
         *
         * @param tableName The name of the table, possibly with a schema
         * @return A pair with the first item as the schema and the second as the table name
         */
        private fun splitSchemaAndTable(tableName: String): Pair<String, String> {
            val parts = tableName.split('.')
            return if (parts.size == 1) Pair("", tableName) else Pair(parts[0], parts[1])
        }

        /**
         * SQL statement to create an index on one or more fields in a JSON document
         *
         * @param tableName The table on which an index should be created (may include schema)
         * @param indexName The name of the index to be created
         * @param fields One or more fields to include in the index
         * @param dialect The SQL dialect to use when creating this index
         * @return A query to create the field index
         */
        fun ensureIndexOn(tableName: String, indexName: String, fields: List<String>, dialect: Dialect): String {
            val (_, tbl) = splitSchemaAndTable(tableName)
            val jsonFields = fields.joinToString(", ") {
                val parts = it.split(' ')
                val direction = if (parts.size > 1) " ${parts[1]}" else ""
                "(" + Field.nameToPath(parts[0], dialect, FieldFormat.SQL) + ") $direction"
            }
            return "CREATE INDEX IF NOT EXISTS idx_${tbl}_$indexName ON $tableName ($jsonFields)"
        }

        /**
         * SQL statement to create a key index for a document table
         *
         * @param tableName The table on which a key index should be created (may include schema)
         * @param dialect The SQL dialect to use when creating this index
         * @return A query to create the key index
         */
        fun ensureKey(tableName: String, dialect: Dialect): String =
            ensureIndexOn(tableName, "key", listOf(Configuration.idField), dialect).replace("INDEX", "UNIQUE INDEX")
    }

    /**
     * Query to insert a document
     *
     * @param tableName The table into which to insert (may include schema)
     * @return A query to insert a document
     */
    fun insert(tableName: String): String =
        "INSERT INTO $tableName VALUES (:data)"

    /**
     * Query to save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
     *
     * @param tableName The table into which to save (may include schema)
     * @return A query to save a document
     */
    fun save(tableName: String): String =
        String.format("INSERT INTO %s VALUES (:data) ON CONFLICT ((data->>'%s')) DO UPDATE SET data = EXCLUDED.data",
            tableName, Configuration.idField)

    /**
     * Query to count documents in a table (this query has no `WHERE` clause)
     *
     * @param tableName The table in which to count documents (may include schema)
     * @return A query to count documents
     */
    fun count(tableName: String): String =
        "SELECT COUNT(*) AS it FROM $tableName"

    /**
     * Query to check for document existence in a table
     *
     * @param tableName The table in which existence should be checked (may include schema)
     * @param where The `WHERE` clause with the existence criteria
     * @return A query to check document existence
     */
    fun exists(tableName: String, where: String): String =
        "SELECT EXISTS (SELECT 1 FROM $tableName WHERE $where) AS it"

    /**
     * Query to select documents from a table (this query has no `WHERE` clause)
     *
     * @param tableName The table from which documents should be found (may include schema)
     * @return A query to retrieve documents
     */
    fun find(tableName: String): String =
        "SELECT data FROM $tableName"

    /**
     * Query to update (replace) a document (this query has no `WHERE` clause)
     *
     * @param tableName The table in which documents should be replaced (may include schema)
     * @return A query to update documents
     */
    fun update(tableName: String): String =
        "UPDATE $tableName SET data = :data"

    /**
     * Query to delete documents from a table (this query has no `WHERE` clause)
     *
     * @param tableName The table in which documents should be deleted (may include schema)
     * @return A query to delete documents
     */
    fun delete(tableName: String): String =
        "DELETE FROM $tableName"

    /**
     * Create an `ORDER BY` clause for the given fields
     *
     * @param fields One or more fields by which to order
     * @param dialect The SQL dialect for the generated clause
     * @return An `ORDER BY` clause for the given fields
     */
    fun orderBy(fields: List<Field<*>>, dialect: Dialect): String {
        if (fields.isEmpty()) return ""
        val orderFields = fields.joinToString(", ") {
            val (field, direction) =
                if (it.name.indexOf(' ') > -1) {
                    val parts = it.name.split(' ')
                    Pair(Field.Named(parts[0]), " " + parts.drop(1).joinToString(" "))
                } else {
                    Pair<Field<*>, String?>(it, null)
                }
            val path =
                if (field.name.startsWith("n:")) {
                    val fld = Field.Named(field.name.substring(2))
                    when (dialect) {
                        Dialect.POSTGRESQL -> "(${fld.path(dialect)})::numeric"
                        Dialect.SQLITE -> fld.path(dialect)
                    }
                } else if (field.name.startsWith("i:")) {
                    val p = Field.Named(field.name.substring(2)).path(dialect)
                    when (dialect) {
                        Dialect.POSTGRESQL -> "LOWER($p)"
                        Dialect.SQLITE -> "$p COLLATE NOCASE"
                    }
                } else {
                    field.path(dialect)
                }
            "$path${direction ?: ""}"
        }
        return "ORDER BY $orderFields"
    }
}
